from django.core.files import File
from avatar.utils import get_default_avatar_url


def string_to_color(username):
    hash = 0
    for ch in username:
        hash = ord(ch) + ((hash << 5) - hash)

    r = str((hash >> (0 * 8)) & 255)
    g = str((hash >> (1 * 8)) & 255)
    b = str((hash >> (2 * 8)) & 255)

    return 'rgb(' + r + ',' + g + ',' + b + ')'


def get_user_avatar_url(user):
    size = 80
    # We use our own method to find the avatar to instead of
    # "get_primary_avatar" as this way we can minimize the reading from disk
    # and set a default thumbnail in case we could not create on.
    # See https://github.com/grantmcconnaughey/django-avatar/pull/187
    avatar = user.avatar_set.order_by("-primary", "-date_uploaded").first()
    if avatar:
        if not avatar.thumbnail_exists(size):
            avatar.create_thumbnail(size)
            # Now check if the thumbnail was actually created
            if not avatar.thumbnail_exists(size):
                # Thumbnail was not saved. There must be some PIL bug
                # with this image type. We store the original file instead.
                avatar.avatar.storage.save(
                    avatar.avatar_name(size),
                    File(avatar.avatar.storage.open(avatar.avatar.name, 'rb'))
                )
        url = avatar.avatar_url(size)
        return {
            'url': url,
            'uploaded': True,
            'html': (
                '<img class="fw-avatar" src="' +
                url +
                '" alt="' + user.username + '">'
            )
        }
    else:
        cl = string_to_color(user.username)
        return {
            'url': get_default_avatar_url(),
            'uploaded': False,
            'html': (
                '<span class="fw-string-avatar" style="background-color: ' +
                cl +
                ';"><span>' +
                user.username[0] +
                '</span></span>'
            )
        }



import time
import multiprocessing
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from testing.testcases import LiveTornadoTestCase
from .editor_helper import EditorHelper
from document.ws_views import WebSocket
from django.conf import settings
import os
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys


class OfflineTests(LiveTornadoTestCase, EditorHelper):
    """
    Tests in which two browsers collaborate and the connection is interrupted.
    """
    user = None
    TEST_TEXT = "Lorem ipsum dolor sit amet."
    MULTILINE_TEST_TEXT = "Lorem ipsum\ndolor sit amet."
    fixtures = [
        'initial_documenttemplates.json',
        'initial_styles.json',
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        driver_data = cls.get_drivers(2)
        cls.driver = driver_data["drivers"][0]
        cls.driver2 = driver_data["drivers"][1]
        cls.client = driver_data["clients"][0]
        cls.client2 = driver_data["clients"][1]
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        cls.driver2.quit()
        super().tearDownClass()

    def setUp(self):
        self.user = self.create_user()
        self.login_user(self.user, self.driver, self.client)
        self.login_user(self.user, self.driver2, self.client2)
        self.doc = self.create_new_document()

    def tearDown(self):
        self.leave_site(self.driver)
        self.leave_site(self.driver2)

    def test_simple(self):
        """
        Test one client going offline in collaborative mode while both clients
        continue to write and whether documents are synched when user returns
        online.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)
        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        p1 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # driver 2 goes offline
        self.driver2.execute_script(
            'window.theApp.page.ws.goOffline()'
        )

        self.driver2.find_element_by_class_name(
            'article-body'
        ).click()

        # Total: 25
        self.driver2.execute_script(
            'window.testCaret.setSelection(25,25)'
        )

        p2 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver2, self.TEST_TEXT)
        )
        p2.start()
        p1.join()
        p2.join()

        # driver 2 goes online
        self.driver2.execute_script(
            'window.theApp.page.ws.goOnline()'
        )

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.TEST_TEXT) * 2,
            len(self.get_contents(self.driver))
        )

        self.assertEqual(
            self.get_contents(self.driver2),
            self.get_contents(self.driver)
        )

    def test_too_many_diffs(self):
        """
        Test one client going offline in collaborative mode while both clients
        continue to write with the connected clients adding too many items to
        the history so that the server no longer can provide it with all
        missing steps. The client therefore needs to recreate the missing steps
        by itself.
        """

        # The history length stored by the server is shortened from 1000 to 1.
        WebSocket.history_length = 1

        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)
        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        p1 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # driver 2 goes offline
        self.driver2.execute_script(
            'window.theApp.page.ws.goOffline()'
        )

        self.driver2.find_element_by_class_name(
            'article-body'
        ).click()

        # Total: 25
        self.driver2.execute_script(
            'window.testCaret.setSelection(25,25)'
        )

        p2 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver2, self.TEST_TEXT)
        )
        p2.start()
        p1.join()
        p2.join()

        # driver 2 goes online
        self.driver2.execute_script(
            'window.theApp.page.ws.goOnline()'
        )

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.TEST_TEXT) * 2,
            len(self.get_contents(self.driver))
        )

        self.assertEqual(
            self.get_contents(self.driver2),
            self.get_contents(self.driver)
        )

        WebSocket.history_length = 1000

    def test_tracking_local_changes(self):
        """
        Test one client going offline in collaborative mode while both clients
        continue to write with the disconnected clients adding enough items to
        the history so that tracking kicks in. The limit of steps is set so
        that tracking kicks in.
        """

        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)
        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        p1 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # driver 2 sets local tracking limit that will be reached
        self.driver2.execute_script(
            'window.theApp.page.mod.collab.doc.merge.trackOfflineLimit = 0'
        )
        # driver 2 sets remote tracking limit that will not be reached
        self.driver2.execute_script((
            'window.theApp.page.mod.collab.doc.merge.'
            'remoteTrackOfflineLimit = 10000'
        ))

        # driver 2 goes offline
        self.driver2.execute_script(
            'window.theApp.page.ws.goOffline()'
        )

        self.driver2.find_element_by_class_name(
            'article-body'
        ).click()

        # Total: 25
        self.driver2.execute_script(
            'window.testCaret.setSelection(25,25)'
        )

        p2 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver2, self.TEST_TEXT)
        )
        p2.start()
        p1.join()
        p2.join()

        # driver 2 goes online
        self.driver2.execute_script(
            'window.theApp.page.ws.goOnline()'
        )

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.TEST_TEXT) * 2,
            len(self.get_contents(self.driver))
        )

        self.assertEqual(
            self.get_contents(self.driver2),
            self.get_contents(self.driver)
        )

        dialogtitle = WebDriverWait(self.driver2, self.wait_time).until(
            EC.element_to_be_clickable((By.CLASS_NAME, "ui-dialog-title"))
        )

        assert dialogtitle.text == 'System message'
        self.driver2.find_element_by_css_selector(
            '.ui-dialog button.fw-orange.fw-button'
        ).click()

        change_tracking_boxes = self.driver2.find_elements_by_css_selector(
            '.margin-box.track'
        )
        self.assertEqual(
            len(change_tracking_boxes),
            1
        )

    def test_tracking_remote_changes(self):
        """
        Test one client going offline in collaborative mode while both clients
        continue to write with the connected clients adding enough items to
        the history so that tracking kicks in. The limit of steps is set so
        that tracking kicks in.
        """

        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)
        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        p1 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # driver 2 sets remote tracking limit that will be reached
        self.driver2.execute_script((
            'window.theApp.page.mod.collab.doc.merge.'
            'remoteTrackOfflineLimit = 0'
        ))
        # driver 2 sets local tracking limit that will not be reached
        self.driver2.execute_script((
            'window.theApp.page.mod.collab.doc.merge.trackOfflineLimit = 10000'
        ))

        # driver 2 goes offline
        self.driver2.execute_script(
            'window.theApp.page.ws.goOffline()'
        )

        self.driver2.find_element_by_class_name(
            'article-body'
        ).click()

        # Total: 25
        self.driver2.execute_script(
            'window.testCaret.setSelection(25,25)'
        )

        p2 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver2, self.TEST_TEXT)
        )
        p2.start()
        p1.join()
        p2.join()

        # driver 2 goes online
        self.driver2.execute_script(
            'window.theApp.page.ws.goOnline()'
        )

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.TEST_TEXT) * 2,
            len(self.get_contents(self.driver))
        )

        self.assertEqual(
            self.get_contents(self.driver2),
            self.get_contents(self.driver)
        )

        dialogtitle = WebDriverWait(self.driver2, self.wait_time).until(
            EC.element_to_be_clickable((By.CLASS_NAME, "ui-dialog-title"))
        )

        assert dialogtitle.text == 'System message'
        self.driver2.find_element_by_css_selector(
            '.ui-dialog button.fw-orange.fw-button'
        ).click()

        change_tracking_boxes = self.driver2.find_elements_by_css_selector(
            '.margin-box.track'
        )
        self.assertEqual(
            len(change_tracking_boxes),
            1
        )

    def test_too_many_diffs_with_tracking(self):
        """
        Test one client going offline in collaborative mode while both clients
        continue to write with the connected clients adding too many items to
        the history so that the server no longer can provide it with all
        missing steps. The client therefore needs to recreate the missing steps
        by itself. The limit of steps is set so that tracking kicks in.
        """

        # The history length stored by the server is shortened from 1000 to 1.
        WebSocket.history_length = 1

        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)
        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        p1 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # driver 2 sets tracking limit
        self.driver2.execute_script(
            'window.theApp.page.mod.collab.doc.merge.trackOfflineLimit = 0'
        )

        # driver 2 goes offline
        self.driver2.execute_script(
            'window.theApp.page.ws.goOffline()'
        )

        self.driver2.find_element_by_class_name(
            'article-body'
        ).click()

        # Total: 25
        self.driver2.execute_script(
            'window.testCaret.setSelection(25,25)'
        )

        p2 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver2, self.TEST_TEXT)
        )
        p2.start()
        p1.join()
        p2.join()

        # driver 2 goes online
        self.driver2.execute_script(
            'window.theApp.page.ws.goOnline()'
        )

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.TEST_TEXT) * 2,
            len(self.get_contents(self.driver))
        )

        self.assertEqual(
            self.get_contents(self.driver2),
            self.get_contents(self.driver)
        )

        dialogtitle = WebDriverWait(self.driver2, self.wait_time).until(
            EC.element_to_be_clickable((By.CLASS_NAME, "ui-dialog-title"))
        )

        assert dialogtitle.text == 'System message'
        self.driver2.find_element_by_css_selector(
            '.ui-dialog button.fw-orange.fw-button'
        ).click()

        change_tracking_boxes = self.driver2.find_elements_by_css_selector(
            '.margin-box.track'
        )
        self.assertEqual(
            len(change_tracking_boxes),
            1
        )

        WebSocket.history_length = 1000

    def test_failed_authentication(self):
        """
        Test One Client Going offline, while the other client is still
        editing the document. The client that goes offline has its
        session expired, while it is offline.
        When it comes back online, the user sees a dialog explaining the
        situation and the offline version of the document is downloaded.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)
        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        p1 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # driver 2 goes offline
        self.driver2.execute_script(
            'window.theApp.page.ws.goOffline()'
        )

        self.driver2.find_element_by_class_name(
            'article-body'
        ).click()

        # Total: 25
        self.driver2.execute_script(
            'window.testCaret.setSelection(25,25)'
        )

        p2 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver2, self.TEST_TEXT)
        )
        p2.start()
        p1.join()
        p2.join()

        # Clear cookie before coming back online
        self.driver2.delete_cookie(settings.SESSION_COOKIE_NAME)

        # driver 2 goes online
        self.driver2.execute_script(
            'window.theApp.page.ws.goOnline()'
        )

        # Check that session expiration dialog is displayed
        element = WebDriverWait(self.driver2, self.wait_time).until(
            EC.visibility_of_element_located(
                (By.ID, 'session_expiration_dialog')
            )
        )
        self.assertEqual(element.is_displayed(), True)

    def test_conflicting_changes(self):
        """
        Test one client going offline, while the other client is still
        editing the document. The client that is offline adds
        content to a paragraph. This paragraph is deleted by the online user.
        Because of this conflict, the merge window opens up.
        """
        # The history length stored by the server is shortened from 1000 to 1.
        WebSocket.history_length = 1

        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)
        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        p1 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # driver 2 goes offline
        self.driver2.execute_script(
            'window.theApp.page.ws.goOffline()'
        )

        self.driver2.find_element_by_class_name(
            'article-body'
        ).click()

        # Start writing text in the middle to cause conflict
        # when online user deletes data.
        self.driver2.execute_script(
            'window.testCaret.setSelection(27,27)'
        )

        p2 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver2, self.TEST_TEXT)
        )
        p2.start()
        p1.join()
        p2.join()

        # Delete all the content from client 1 to cause conflict.
        for i in range(0, len(self.TEST_TEXT)):
            actions = ActionChains(self.driver)
            actions.send_keys(Keys.BACKSPACE)
            actions.perform()

        # Driver 2 goes online
        self.driver2.execute_script(
            'window.theApp.page.ws.goOnline()'
        )

        # Check whether the merge window is available in driver2
        element = WebDriverWait(self.driver2, self.wait_time).until(
            EC.visibility_of_element_located(
                (By.ID, 'editor-merge-view')
            )
        )
        self.assertEqual(element.is_displayed(), True)
        # Check that the documents in main editors are synced!
        self.assertEqual(
            self.get_contents(self.driver2),
            self.get_contents(self.driver)
        )

        # Change the websocket history length back to its original value
        WebSocket.history_length = 1000


class FunctionalOfflineTests(LiveTornadoTestCase, EditorHelper):
    """
    Tests in which one user works offline. The Service Worker is
    also installed in these tests.
    """
    user = None
    TEST_TEXT = "Lorem ipsum dolor sit amet."
    MULTILINE_TEST_TEXT = "Lorem ipsum\ndolor sit amet."
    fixtures = [
        'initial_documenttemplates.json',
        'initial_styles.json',
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        driver_data = cls.get_drivers(1)
        cls.driver = driver_data["drivers"][0]
        cls.client = driver_data["clients"][0]
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        super().tearDownClass()

    def setUp(self):
        self.user = self.create_user()
        self.login_user(self.user, self.driver, self.client)
        self.driver.execute_script('window.theApp.installServiceWorker()')
        self.doc = self.create_new_document()

    def tearDown(self):
        self.leave_site(self.driver)

    def test_service_workers(self):
        """
        Test one client going offline after writing some text and inserting
        some images. While offline client tries to export to HTML
        and prints the PDF of the document.
        """
        self.load_document_editor(self.driver, self.doc)

        self.add_title(self.driver)
        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        self.type_text(self.driver, self.TEST_TEXT)

        # We add a figure
        button = self.driver.find_element_by_xpath('//*[@title="Figure"]')
        button.click()

        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, "span.math-field")
            )
        )
        self.driver.find_element_by_css_selector(
            "div.figure-category"
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Photo"]'
        ).click()

        # click on 'Insert image' button
        self.driver.find_element_by_id('insert-figure-image').click()

        upload_button = WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (
                    By.XPATH,
                    '//*[normalize-space()="Add new image"]'
                )
            )
        )
        upload_button.click()

        # image path
        image_path = os.path.join(
            settings.PROJECT_PATH,
            'document/tests/uploads/image.png'
        )

        # in order to select the image we send the image path in the
        # LOCAL MACHINE to the input tag
        upload_image_url = WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.XPATH, '//*[@id="editimage"]/div[1]/input[2]')
            )
        )
        upload_image_url.send_keys(image_path)

        # click on 'Upload' button
        self.driver.find_element_by_xpath(
            '//*[contains(@class, "ui-button") and normalize-space()="Upload"]'
        ).click()

        # click on 'Use image' button
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, '.fw-data-table i.fa-check')
            )
        )

        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Use image"]'
        ).click()
        self.driver.find_element_by_css_selector("button.fw-dark").click()

        caption = WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, "div.article-body figure figcaption")
            )
        )

        caption.click()

        caption.send_keys("Caption")

        ActionChains(self.driver).send_keys(
            Keys.RIGHT
        ).perform()

        # driver goes offline
        self.driver.execute_script(
            'window.theApp.page.ws.goOffline()'
        )

        # Check that the html export works fine!
        # Click on the menu
        self.driver.find_element_by_xpath(
            "//span[contains(@title,'Export of the document contents')]"
        ).click()

        # Click on the HTML export
        self.driver.find_element_by_xpath(
            "//span[contains(@title,'Export the document to an HTML file.')]"
        ).click()

        # Check that the alert box is displayed.
        alert_element = WebDriverWait(self.driver, self.wait_time).until(
            EC.visibility_of_element_located(
                (By.CLASS_NAME, 'alerts-info')
            )
        )
        self.assertEqual(alert_element.is_displayed(), True)
        time.sleep(1)
        # Check the same for PDF export too !
        # Click on the file menu
        self.driver.find_element_by_xpath(
            "//span[contains(@title,'File handling')]"
        ).click()
        time.sleep(1)

        # Click on the Print PDF button
        self.driver.find_element_by_xpath(
            "//span[contains(@title,'Either print or" +
            " create a PDF using your browser print dialog.')]"
        ).click()

        # Check that the alert box is displayed.
        alert_element = WebDriverWait(self.driver, self.wait_time).until(
            EC.visibility_of_element_located(
                (By.CLASS_NAME, 'alerts-info')
            )
        )
        self.assertEqual(alert_element.is_displayed(), True)

    def test_disabled_options(self):
        """
        Test one user going offline after writing some text.
        While the user is offline, he tries different export
        options which are disabled. He wants to upload an image,
        but the button is gone.
        """
        self.load_document_editor(self.driver, self.doc)
        self.add_title(self.driver)

        # driver goes offline
        self.driver.execute_script(
            'window.theApp.page.ws.goOffline()'
        )
        self.driver.execute_script(
            'window.theApp.ws.goOffline()'
        )

        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        # Type some text
        self.type_text(self.driver, self.TEST_TEXT)

        # Check the share and create revision buttons are disabled.
        file_menu = self.driver.find_element_by_xpath(
            "//span[contains(@title,'File handling')]"
        )
        file_menu.click()
        time.sleep(1)

        share_button = self.driver.find_element_by_xpath(
            "//span[contains(@title,'Share the document with other users.')]"
        )
        share_button_classes = share_button.get_attribute("class").split(' ')
        self.assertEqual('disabled' in share_button_classes, True)

        save_revision_button = self.driver.find_element_by_xpath(
            "//span[contains(@title,'Save a revision of " +
            "the current document.')]"
        )
        save_revision_button_classes = save_revision_button.get_attribute(
            "class"
        ).split(' ')
        self.assertEqual('disabled' in save_revision_button_classes, True)

        # Check that the EPUB, LaTex and JATS exports are disabled
        # when user is offline.
        export_menu = self.driver.find_element_by_xpath(
            "//span[contains(@title,'Export of the document contents')]"
        )
        export_menu.click()

        epub_export_button = self.driver.find_element_by_xpath(
            "//span[contains(@title,'Export the document to " +
            "an Epub electronic reader file.')]"
        )
        epub_export_button_classes = epub_export_button.get_attribute(
            "class"
        ).split(' ')
        self.assertEqual('disabled' in epub_export_button_classes, True)

        latex_export_button = self.driver.find_element_by_xpath(
            "//span[contains(@title,'Export the document to an LaTeX file.')]"
        )
        latex_export_button_classes = latex_export_button.get_attribute(
            "class"
        ).split(' ')
        self.assertEqual('disabled' in latex_export_button_classes, True)

        jats_export_button = self.driver.find_element_by_xpath(
            "//span[contains(@title,'Export the document to a Journal" +
            " Archiving and Interchange Tag Library NISO JATS Version 1.2 " +
            "file.')]"
        )
        jats_export_button_classes = jats_export_button.get_attribute(
            "class"
        ).split(' ')
        self.assertEqual('disabled' in jats_export_button_classes, True)

        # Check that the Switching between styles is disabled.
        settings_menu = self.driver.find_element_by_xpath(
            "//span[contains(@title,'Configure settings of this document.')]"
        )
        settings_menu.click()

        doc_style_button = self.driver.find_element_by_xpath(
            "//span[contains(@title,'Choose your preferred document style.')]"
        )
        doc_style_button_classes = doc_style_button.get_attribute(
            "class"
        ).split(' ')
        self.assertEqual('disabled' in doc_style_button_classes, True)

        # Try to upload a figure
        button = self.driver.find_element_by_xpath('//*[@title="Figure"]')
        button.click()

        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, "span.math-field")
            )
        )
        self.driver.find_element_by_css_selector(
            "div.figure-category"
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Photo"]'
        ).click()

        # click on 'Insert image' button
        self.driver.find_element_by_id('insert-figure-image').click()

        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (
                    By.XPATH,
                    '//*[normalize-space()="Use image"]'
                )
            )
        )

        upload_buttons = self.driver.find_elements_by_xpath(
            '//*[normalize-space()="Add new image"]'
        )
        self.assertEqual(
            len(upload_buttons),
            0
        )

        # Come back online to prevent pop up from showing up.
        self.driver.execute_script(
            'window.theApp.page.ws.goOnline()'
        )
        self.driver.execute_script(
            'window.theApp.ws.goOnline()'
        )
        WebDriverWait(self.driver, self.wait_time).until(
            lambda driver: driver.execute_script(
                'return window.theApp.ws.connected'
            )
        )

    def test_indexedDB(self):
        """
        Testing a user going offline after logging in.
        Test that the documents overview page is
        rendered from indexed DB when the user is
        offline.
        """
        # Load the documents overview page
        self.driver.get(self.live_server_url)

        # Go offline
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'fw-contents'))
        )
        self.driver.execute_script('window.theApp.ws.goOffline()')

        # Click the documents overview page button to see
        # if it's loaded from indexed DB
        doc_overview_menu = self.driver.find_element_by_xpath(
            "//a[contains(@title,'edit documents')]"
        )
        doc_overview_menu.click()

        # Check the document table is rendered even when offline!
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'fw-contents'))
        )
        doc_row = self.driver.find_element_by_xpath(
            "//a[@href='/document/"+str(self.doc.id)+"/']"
        )
        self.assertEqual(doc_row.is_displayed(), True)

        # Check that the alert regarding offline is shown.
        alert_element = self.driver.find_element_by_class_name('alerts-info')
        self.assertEqual(alert_element.is_displayed(), True)



from builtins import range
from builtins import object
import re
import os

from django.test import Client
from selenium import webdriver
from allauth.account.models import EmailAddress
from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import User


class SeleniumHelper(object):
    """
    Methods for manipulating django and the browser for testing purposes.
    """

    @classmethod
    def get_drivers(cls, number, download_dir=False, user_agent=False):
        # django native clients, to be used for faster login.
        clients = []
        for i in range(number):
            clients.append(Client())
        drivers = []
        wait_time = 0
        options = webdriver.ChromeOptions()
        if download_dir:
            prefs = {
                "download.default_directory": download_dir,
                "download.prompt_for_download": False,
                "download.directory_upgrade": True
            }
            options.add_experimental_option("prefs", prefs)
        if user_agent:
            options.add_argument("user-agent={}".format(user_agent))
        if os.getenv("CI"):
            options.binary_location = '/usr/bin/google-chrome-stable'
            options.add_argument('--headless')
            options.add_argument('--disable-gpu')
            chromedriver_filename = None
            wait_time = 10
        else:
            from chromedriver_binary import chromedriver_filename
            wait_time = 6
        for i in range(number):
            if chromedriver_filename:
                driver = webdriver.Chrome(
                    chromedriver_filename,
                    options=options
                )
            else:
                driver = webdriver.Chrome(
                    options=options
                )
            drivers.append(driver)
        for driver in drivers:
            # Set sizes of browsers so that all buttons are visible.
            driver.set_window_position(0, 0)
            driver.set_window_size(1920, 1080)
        return {
            "clients": clients,
            "drivers": drivers,
            "wait_time": wait_time
        }

    def find_urls(self, string):
        return re.findall(
            (
                'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*(),]|'
                '(?:%[0-9a-fA-F][0-9a-fA-F]))+'
            ),
            string
        )

    # create django data
    def create_user(
        self,
        username='User',
        email='test@example.com',
        passtext='p4ssw0rd'
    ):
        user = User.objects.create(
            username=username,
            email=email,
            password=make_password(passtext),
            is_active=True
        )
        user.save()

        # avoid the unverified-email login trap
        EmailAddress.objects.create(
            user=user,
            email=email,
            verified=True,
            primary=True
        ).save()

        return user

    # drive browser
    def login_user(self, user, driver, client):
        client.force_login(user=user)
        cookie = client.cookies['sessionid']
        if driver.current_url == 'data:,':
            # To set the cookie at the right domain we load the front page.
            driver.get('%s%s' % (self.live_server_url, '/'))
        driver.add_cookie({
            'name': 'sessionid',
            'value': cookie.value,
            'secure': False,
            'path': '/'
        })

    def leave_site(self, driver):
        driver.get('data:,')



from base.management import BaseCommand
import zipfile
import os
import shutil
import magic

from django.conf import settings


def zip_folder(path, zip_file):
    file_paths = []
    for root, dirs, files in os.walk(path):
        for file in files:
            out_file = os.path.join(root, file)
            relative_out_file = os.path.relpath(out_file, path)
            zip_file.write(out_file, relative_out_file)
            mimetype = magic.from_file(out_file, mime=True)
            # Override mimetype for CSS files
            if out_file.split('.')[-1].lower() == 'css':
                mimetype = 'text/css'
            file_paths.append({
                'path': relative_out_file,
                'mimetype': mimetype
            })
    return file_paths


def opf_entries(file_paths):
    opf_text = (
        '// This file is auto-generated. CHANGES WILL BE OVERWRITTEN! '
        'Re-generate by running ./manage.py bundle_mathlive.\n'
    )
    opf_text += 'export const mathliveOpfIncludes = `\n'
    for index, file_path in enumerate(file_paths):
        mimetype = file_path['mimetype']
        path = file_path['path']
        if path.split('.')[-1] == 'woff':
            mimetype = 'font/woff'
        elif path.split('.')[-1] == 'woff2':
            mimetype = 'font/woff2'
        opf_text += \
            '<item id="mathlive-%d" href="css/%s" media-type="%s" />\n' % (
                index,
                path,
                mimetype
            )
    opf_text += '`'
    return opf_text


class Command(BaseCommand):
    args = ''
    help = (
        'Create a zip file containing the mathlive style files to be bundled '
        'with EPUB and HTML exports'
    )

    def handle(self, *args, **options):
        self.stdout.write("Bundling MathLive")
        # Copy MathLive CSS
        mathlive_css_path = os.path.join(
            settings.PROJECT_PATH, "static-libs/css/libs/mathlive/")
        if not os.path.exists(mathlive_css_path):
            os.makedirs(mathlive_css_path)
        with open(os.path.join(
            settings.PROJECT_PATH,
            ".transpile/node_modules/mathlive/dist/mathlive-static.css"
        )) as f:
            with open(os.path.join(
                mathlive_css_path,
                "mathlive.css"
            ), 'w') as g:
                g.write(f.read().replace('url(fonts/', 'url(media/'))
        mathlive_fonts_path = os.path.join(
            settings.PROJECT_PATH, "static-libs/css/libs/mathlive/media/")
        if not os.path.exists(mathlive_fonts_path):
            os.makedirs(mathlive_fonts_path)
        fontfiles_src_path = os.path.join(
            settings.PROJECT_PATH,
            ".transpile/node_modules/mathlive/dist/fonts/"
        )
        fontfiles = os.listdir(fontfiles_src_path)
        for filename in fontfiles:
            if filename[0] == '.':
                continue
            full_filename = os.path.join(fontfiles_src_path, filename)
            if os.path.isfile(full_filename):
                shutil.copy(full_filename, mathlive_fonts_path)
        zip_file_path = os.path.join(
            settings.PROJECT_PATH, 'static-libs/zip/mathlive_style.zip')
        zip_dir = os.path.dirname(zip_file_path)
        if not os.path.exists(zip_dir):
            os.makedirs(zip_dir)
        if os.path.exists(zip_file_path):
            os.remove(zip_file_path)
        zip_file = zipfile.ZipFile(zip_file_path, 'w', zipfile.ZIP_DEFLATED)
        in_dir = os.path.dirname(mathlive_css_path)
        file_paths = zip_folder(in_dir, zip_file)
        zip_file.close()

        opf_file_contents = opf_entries(file_paths)
        opf_file_path = os.path.join(
            settings.PROJECT_PATH,
            'static-libs/js/modules/mathlive/opf_includes.js'
        )
        opf_dir = os.path.dirname(opf_file_path)
        if not os.path.exists(opf_dir):
            os.makedirs(opf_dir)
        opf_file = open(opf_file_path, "wb")
        opf_file.write(opf_file_contents.encode('utf8'))
        opf_file.close()



from django.core.management.base import BaseCommand
import os
import subprocess
from tempfile import mkstemp

from django.conf import settings
from django.core.management import call_command


class Command(BaseCommand):
    help = (
        'Copy the document schema used in the frontend to a JSON file so that '
        'also the backend can make use of it.'
    )

    def handle(self, *args, **options):
        self.stdout.write("Copying document schema")
        call_command("npm_install")
        json_path = os.path.join(
            settings.PROJECT_PATH,
            "static-libs/json/"
        )
        if not os.path.exists(json_path):
            os.makedirs(json_path)
        schema_json_path = os.path.join(
            json_path,
            "schema.json"
        )
        js_file = open(os.path.join(
            settings.PROJECT_PATH,
            "static-transpile/js/schema_export.js"
        ))
        js = js_file.read()
        js_file.close()
        node_file, node_file_path = mkstemp()
        with open(node_file, 'w') as f:
            f.write('global.window = {}; global.gettext = ()=>{};\n' + js)
        json = subprocess.check_output(["node", node_file_path]).decode(
            'utf-8'
        )
        os.unlink(node_file_path)
        schema_json = open(schema_json_path, 'w')
        schema_json.write(json)
        schema_json.close()



from django.conf.urls import url

from . import views

urlpatterns = [
    url('^save/$', views.save, name='bibliography_save'),
    url('^delete/$', views.delete, name='bibliography_delete'),
    url(
        '^save_category/$',
        views.save_category,
        name='bibliopgraphy_save_category'
    ),
    url(
        '^delete_category/$',
        views.delete_category,
        name='bibliography_delete_category'
    ),
    url('^biblist/$', views.biblist, name='bibliography_biblist')
]



from django.contrib import admin

from .models import UserProfile, TeamMember


class UserProfileAdmin(admin.ModelAdmin):
    pass


admin.site.register(UserProfile, UserProfileAdmin)


class TeamMemberAdmin(admin.ModelAdmin):
    pass


admin.site.register(TeamMember, TeamMemberAdmin)



from django.contrib import admin

from . import models


class DocumentStyleFileAdmin(admin.ModelAdmin):
    pass


admin.site.register(models.DocumentStyleFile, DocumentStyleFileAdmin)


class DocumentStyleFileInline(admin.TabularInline):
    model = models.DocumentStyleFile
    extra = 1


class DocumentStyleAdmin(admin.ModelAdmin):
    inlines = [DocumentStyleFileInline, ]


admin.site.register(models.DocumentStyle, DocumentStyleAdmin)


class ExportTemplateAdmin(admin.ModelAdmin):
    pass


admin.site.register(models.ExportTemplate, ExportTemplateAdmin)



# Generated by Django 2.2.9 on 2020-02-19 13:52

from django.db import migrations, models
import django.db.models.deletion
import style.models


class Migration(migrations.Migration):

    replaces = [('style', '0001_squashed_0002_auto_20151226_1110'), ('style', '0002_auto_20190316_0241'), ('style', '0003_auto_20190803_0749'), ('style', '0004_auto_20190809_1655'), ('style', '0005_documentstyle_documentstylefile_exporttemplate'), ('style', '0006_auto_20190809_1757'), ('style', '0007_auto_20190811_1204'), ('style', '0008_auto_20190830_0627')]

    initial = True

    dependencies = [
        ('document', '0001_squashed_20200219'),
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentStyle',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(default='Default', help_text='The human readable title.', max_length=128)),
                ('slug', models.SlugField(default='default', help_text='The base of the filenames the style occupies.', max_length=20)),
                ('contents', models.TextField(default='', help_text='The CSS style definiton.')),
                ('document_template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='document.DocumentTemplate')),
            ],
            options={
                'unique_together': {('slug', 'document_template')},
            },
        ),
        migrations.CreateModel(
            name='ExportTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(default='Default', help_text='The human readable title.', max_length=128)),
                ('file_type', models.CharField(choices=[('docx', 'DOCX'), ('odt', 'ODT')], max_length=5)),
                ('template_file', models.FileField(upload_to=style.models.template_filename)),
                ('document_template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='document.DocumentTemplate')),
            ],
            options={
                'unique_together': {('title', 'document_template')},
            },
        ),
        migrations.CreateModel(
            name='DocumentStyleFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('file', models.FileField(help_text='A file references in the style. The filename will be replaced with the final url of the file in the style.', upload_to=style.models.documentstylefile_location)),
                ('filename', models.CharField(help_text='The original filename.', max_length=255)),
                ('style', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='style.DocumentStyle')),
            ],
            options={
                'unique_together': {('filename', 'style')},
            },
        ),
    ]



import os
import distutils
import setuptools
from subprocess import call
from glob import glob
from setuptools.command.sdist import sdist as _sdist
from setuptools.command.bdist_egg import bdist_egg as _bdist_egg
from setuptools.command.install import install as _install
from setuptools.command.build_py import build_py as _build_py


def read(name):
    with open(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), name)
    ) as f:
        return f.read()


class compilemessages(distutils.cmd.Command):
    """A custom command to create *.mo files for each language."""
    description = "Create *.mo files to be included in the final package."
    user_options = []

    def initialize_options(self):
        """Set default values for options."""
        pass

    def finalize_options(self):
        """Post-process options."""
        pass

    def run(self):
        import subprocess
        from pathlib import Path
        for path in Path(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                'fiduswriter/locale'
            )
        ).rglob('*.po'):
            command = [
                'msgfmt',
                '-o',
                path.name.replace('.po', '.mo'),
                path.name
            ]
            self.announce(
                'Running command: %s in %s' % (
                    str(' '.join(command)),
                    path.parent
                ),
                level=distutils.log.INFO
            )
            subprocess.check_call(command, cwd=path.parent)

# From https://github.com/pypa/setuptools/pull/1574
class build_py(_build_py):
    def find_package_modules(self, package, package_dir):
        modules = super().find_package_modules(package, package_dir)
        patterns = self._get_platform_patterns(
            self.exclude_package_data,
            package,
            package_dir,
        )

        excluded_module_files = []
        for pattern in patterns:
            excluded_module_files.extend(glob(pattern))

        for f in excluded_module_files:
            for module in modules:
                if module[2] == f:
                    modules.remove(module)
        return modules



class install(_install):
    def run(self):
        call(["pip install wheel --no-clean"], shell=True)
        # From https://stackoverflow.com/questions/21915469/python-setuptoo
        # ls-install-requires-is-ignored-when-overriding-cmdclass
        if self.old_and_unmanageable or self.single_version_externally_managed:
            return _install.run(self)

        # Attempt to detect whether we were called from setup() or by another
        # command.  If we were called by setup(), our caller will be the
        # 'run_command' method in 'distutils.dist', and *its* caller will be
        # the 'run_commands' method.  If we were called any other way, our
        # immediate caller *might* be 'run_command', but it won't have been
        # called by 'run_commands'.  This is slightly kludgy, but seems to
        # work.
        #
        caller = sys._getframe(2)
        caller_module = caller.f_globals.get('__name__','')
        caller_name = caller.f_code.co_name

        if caller_module != 'distutils.dist' or caller_name != 'run_commands':
            # We weren't called from the command line or setup(), so we
            # should run in backward-compatibility mode to support bdist_*
            # commands.
            _install.run(self)
        else:
            self.do_egg_install()


class bdist_egg(_bdist_egg):
    def run(self):
        self.run_command('compilemessages')
        _bdist_egg.run(self)


class sdist(_sdist):
    """Custom build command."""

    def run(self):
        self.run_command('compilemessages')
        _sdist.run(self)

cmdclass = {
    'compilemessages': compilemessages,
    'sdist': sdist,
    'build_py': build_py,
    'bdist_egg': bdist_egg,
    'install': install
}

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
    class bdist_wheel(_bdist_wheel):
        """Custom build command."""

        def run(self):
            self.run_command('compilemessages')
            _bdist_wheel.run(self)
    cmdclass['bdist_wheel'] = bdist_wheel
except ImportError:
    pass


setuptools.setup(
    cmdclass=cmdclass,
    name="fiduswriter",
    version=read('fiduswriter/version.txt').splitlines()[0],
    description="A semantic wordprocessor for academic purposes",
    license="AGPL",
    author="Lund Info AB",
    author_email="mail@lundinfo.com",
    url="https://www.fiduswriter.org",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Framework :: Django :: 3.1",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
        "Topic :: Text Editors :: Word Processors",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Text Processing :: Markup :: LaTeX",
        "Topic :: Text Processing :: Markup :: XML",
        "Topic :: Utilities",
    ],
    packages=setuptools.find_namespace_packages(include=['fiduswriter']),
    include_package_data=True,
    exclude_package_data={
        "": [
            "travis/*",
            "build/*",
            "fiduswriter/media/*",
            "fiduswriter/.transpile/*",
            "fiduswriter/static-transpile/*",
            "fiduswriter/static-collected/*",
            "fiduswriter/static-libs/*",
            "fiduswriter/venv/*",
            "fiduswriter/media/*",
            "fiduswriter/static-transpile/*",
            "fiduswriter/static-libs/*",
            "fiduswriter/book/*",
            "fiduswriter/citation-api-import/*",
            "fiduswriter/languagetool/*",
            "fiduswriter/npm_mjs/*",
            "fiduswriter/ojs/*",
            "fiduswriter/phplist/*",
            "fiduswriter/payment/*"
        ]
    },
    python_requires='>=3',
    install_requires=read('fiduswriter/requirements.txt').splitlines(),
    extras_require={
        "books": "fiduswriter-books ~= 3.9.2",
        "citation-api-import": "fiduswriter-citation-api-import ~= 3.9.0",
        "languagetool": "fiduswriter-languagetool ~= 3.9.1",
        "ojs": "fiduswriter-ojs ~= 3.9.0",
        "phplist": "fiduswriter-phplist ~= 3.9.0",
        "github-export": "fiduswriter-github-export ~= 3.9.0",
        "mysql": read('fiduswriter/mysql-requirements.txt').splitlines(),
        "postgresql": read(
            'fiduswriter/postgresql-requirements.txt'
        ).splitlines()
    },
    entry_points={
        "console_scripts": [
            "fiduswriter=fiduswriter.manage:entry"
        ]
    }
)



from django.conf.urls import include, url
from . import views


urlpatterns = [
    url('^save/$', views.save_profile, name='save_profile'),

    # Show user profiles
    url(
        '^team/list/$',
        views.list_team_members,
        name='list_team_members'
    ),
    # Outdated but we need it to allow for update of 3.7 instances.
    # Can be removed in 3.9.
    url(
        '^info/$',
        views.info,
        name='info'
    ),
    url('^avatar/delete/$', views.delete_avatar, name="delete_avatar"),
    url('^avatar/upload/$', views.upload_avatar, name="upload_avatar"),
    url(
        '^passwordchange/$',
        views.password_change,
        name="password_change"
    ),
    url('^emailadd/$', views.add_email, name="add_email"),
    url('^emaildelete/$', views.delete_email, name="delete_email"),
    url('^emailprimary/$', views.primary_email, name="primary_email"),

    url(
        '^socialaccountdelete/$',
        views.delete_socialaccount,
        name="delete_socialaccount"
    ),
    # Delete a user profile
    url('^delete/$', views.delete_user, name='delete_user'),

    url(
        '^teammember/add',
        views.add_team_member,
        name='add_team_member'
    ),
    url(
        '^teammember/remove',
        views.remove_team_member,
        name='remove_team_member'
    ),

    # User avatar handling
    url('^avatar/', include('avatar.urls')),

    url(
        "^get_confirmkey_data/$",
        views.get_confirmkey_data,
        name="get_confirmkey_data"
    ),

    url(r"^signup/$", views.signup, name="account_signup"),

    # Authentication handling
    url('', include('allauth.urls')),
]



# Generated by Django 3.1 on 2020-08-20 03:10

from django.db import migrations, models
import usermedia.models


class Migration(migrations.Migration):

    dependencies = [
        ('usermedia', '0005_move_json_data'),
    ]

    operations = [
        migrations.AlterField(
            model_name='documentimage',
            name='copyright',
            field=models.JSONField(default=usermedia.models.default_copyright),
        ),
        migrations.AlterField(
            model_name='userimage',
            name='copyright',
            field=models.JSONField(default=usermedia.models.default_copyright),
        ),
    ]
